using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Collections;
using System.Diagnostics;

namespace FolderCp
{
	internal class DiffResult
	{
		internal List<string> CommonFiles;
		internal List<string> SrcOnlyFiles;
		internal List<string> DstOnlyFiles;

		internal List<string> CommonDirs;
		internal List<string> SrcOnlyDirs;
		internal List<string> DstOnlyDirs;

		internal DiffResult()
		{
			CommonFiles = new List<string>();
			SrcOnlyFiles = new List<string>();
			DstOnlyFiles = new List<string>();

			CommonDirs = new List<string>();
			SrcOnlyDirs = new List<string>();
			DstOnlyDirs = new List<string>();
		}
	}

	internal class Folder
	{
		internal string Name;
		internal string AbsPath;
		internal string RelPath;

		internal List<string> Files;
		internal Dictionary<string, Folder> Dirs;

		internal Folder(Folder parent, string dirname, string path)
		{
			if (parent == null)
			{
				Name = "(root)";
				RelPath = Path.DirectorySeparatorChar.ToString();
			}
			else
			{
				Name = dirname;
				RelPath = Path.Combine(parent.RelPath, dirname);
			}

			AbsPath = path;
			DirectoryInfo cur_di = new DirectoryInfo(path);

			// filtering files
			Files = new List<string>();
			foreach (FileInfo fi in cur_di.GetFiles())
			{
				bool bExclude = false;

				foreach (CompiledFilter cf in MainForm.CompiledFilters)
				{
					if (cf.Match( Path.Combine(RelPath, fi.Name) ))
					{
						bExclude = cf.IsExclude;
						break;
					}
				}

				if (bExclude)
				{
					MainForm.s_this.Invoke(MainForm.s_this.dele_ReportFile,
						new Object[] { fi.FullName, fi.Length, ReportCategory.EXCLUDE });
				}
				else
				{
					Files.Add(fi.Name);
				}
			}

			// filtering directories
			Dirs = new Dictionary<string, Folder>();
			foreach (DirectoryInfo di in cur_di.GetDirectories())
			{
				bool bExclude = false;

				foreach (CompiledFilter cf in MainForm.CompiledFilters)
				{
					if (cf.Match(Path.Combine(RelPath, di.Name)))
					{
						bExclude = cf.IsExclude;
						break;
					}
				}

				if (bExclude)
				{
					MainForm.s_this.Invoke(MainForm.s_this.dele_ReportFile,
						new Object[] { di.FullName, -1, ReportCategory.EXCLUDE });
				}
				else
				{
					Folder sub = new Folder(this, di.Name, di.FullName);
					if (sub.Files.Count > 0 || sub.Dirs.Count > 0)
						Dirs[di.Name] = sub;
				}
			}
		}

		internal int GetFileCount()
		{
			int cnt = Files.Count;

			foreach (Folder sub in Dirs.Values)
			{
				cnt += sub.GetFileCount();
			}		

			return cnt;
		}

		internal DiffResult MakeDiff(string dstpath)
		{
			DiffResult ret = new DiffResult();

			DirectoryInfo di = new DirectoryInfo(dstpath);
			
			FileInfo[] dfiles = di.GetFiles();
			if (Files.Count == 0)
			{
				// no source files
				foreach (FileInfo f in dfiles)
					ret.DstOnlyFiles.Add(f.Name);
			}
			else
			{
				bool[] common = new bool[Files.Count];
				
				foreach (FileInfo f in dfiles)
				{
					bool isCommon = false;
					for(int i=0; i<Files.Count; i++)
					{
						if(common[i] == false && String.Compare(Files[i], f.Name, true) == 0)
						{
							common[i] = true;
							isCommon = true;
							ret.CommonFiles.Add(f.Name);
							break;
						}
					}

					if(!isCommon)
					{
						ret.DstOnlyFiles.Add(f.Name);
					}
				}

				for(int i=0; i<Files.Count; i++)
				{
					if (common[i] == false)
						ret.SrcOnlyFiles.Add(Files[i]);
				}
			}

			DirectoryInfo[] ddirs = di.GetDirectories();
			if (Dirs.Count == 0)
			{
				// no source directories
				foreach (DirectoryInfo d in ddirs)
					ret.DstOnlyDirs.Add(d.Name);
			}
			else
			{
				Dictionary<string,bool> common = new Dictionary<string,bool>();

				foreach (DirectoryInfo d in ddirs)
				{
					bool isCommon = false;
					foreach(string key in Dirs.Keys)					
					{
						if (common.ContainsKey(key) == false 
						&& String.Compare(key, d.Name, true) == 0)
						{
							common[key] = true;
							isCommon = true;
							ret.CommonDirs.Add(key);
							break;
						}
					}

					if (!isCommon)
					{
						ret.DstOnlyDirs.Add(d.Name);
					}
				}

				foreach (string key in Dirs.Keys)
				{
					if (common.ContainsKey(key) == false)
						ret.SrcOnlyDirs.Add(Dirs[key].Name);
				}
			}

			return ret;
		}
	}
}
